@@ -19,20 +19,21 @@ extern "C" {
 struct nn_rbufpool;
 struct nn_rsample_info;
 struct nn_rdata;
+struct nn_rmsg;
 struct ddsi_tran_listener;
 struct recv_thread_arg;
 struct writer;
 struct proxy_reader;
 
 struct nn_gap_info {
-  int64_t gapstart;
-  int64_t gapend;
+  seqno_t gapstart; // == 0 on init, indicating no gap recorded yet
+  seqno_t gapend;   // >= gapstart
   uint32_t gapnumbits;
   uint32_t gapbits[256 / 32];
 };
 
 void nn_gap_info_init(struct nn_gap_info *gi);
-void nn_gap_info_update(struct ddsi_domaingv *gv, struct nn_gap_info *gi, int64_t seqnr);
+void nn_gap_info_update(struct ddsi_domaingv *gv, struct nn_gap_info *gi, seqno_t seqnr);
 struct nn_xmsg * nn_gap_info_create_gap(struct writer *wr, struct proxy_reader *prd, struct nn_gap_info *gi);
 
 void trigger_recv_threads (const struct ddsi_domaingv *gv);
@@ -41,6 +42,8 @@ uint32_t listen_thread (struct ddsi_tran_listener * listener);
 int user_dqueue_handler (const struct nn_rsample_info *sampleinfo, const struct nn_rdata *fragchain, const ddsi_guid_t *rdguid, void *qarg);
 int add_Gap (struct nn_xmsg *msg, struct writer *wr, struct proxy_reader *prd, seqno_t start, seqno_t base, uint32_t numbits, const uint32_t *bits);
 
+DDS_EXPORT void ddsi_handle_rtps_message (struct thread_state1 * const ts1, struct ddsi_domaingv *gv, ddsi_tran_conn_t conn, const ddsi_guid_prefix_t *guidprefix, struct nn_rbufpool *rbpool, struct nn_rmsg *rmsg, size_t sz, unsigned char *msg, const ddsi_locator_t *srcloc);
+
 #if defined (__cplusplus)
 }
 #endif
