@@ -24,7 +24,7 @@
 #include "dds/ddsi/q_thread.h"
 #include "dds/ddsi/ddsi_threadmon.h"
 #include "dds/ddsi/q_log.h"
-#include "dds/ddsi/q_config.h"
+#include "dds/ddsi/ddsi_config_impl.h"
 #include "dds/ddsi/ddsi_domaingv.h"
 #include "dds/ddsi/sysdeps.h"
 
