@@ -18,7 +18,7 @@
 #include "dds/ddsi/ddsi_domaingv.h"
 #include "dds/ddsi/ddsi_udp.h"
 #include "dds/ddsi/ddsi_tcp.h"
-#include "dds/ddsi/q_config.h"
+#include "dds/ddsi/ddsi_config_impl.h"
 #include "dds/ddsi/q_rtps.h"
 #include "CUnit/Theory.h"
 
