@@ -18,6 +18,7 @@
 #include "idl/string.h"
 #include "symbol.h"
 #include "scope.h"
+#include "tree.h"
 
 static idl_retcode_t
 create_declaration(
@@ -67,7 +68,7 @@ idl_create_scope(
   idl_pstate_t *pstate,
   enum idl_scope_kind kind,
   const idl_name_t *name,
-  const void *node,
+  void *node,
   idl_scope_t **scopep)
 {
   idl_scope_t *scope;
@@ -152,6 +153,12 @@ is_consistent(
   return idl_mask(lhs) == idl_mask(rhs);
 }
 
+static bool
+is_same_type(const void *lhs, const void *rhs)
+{
+  return (idl_mask(lhs) & ~IDL_FORWARD) == (idl_type(rhs) & ~IDL_FORWARD);
+}
+
 idl_retcode_t
 idl_declare(
   idl_pstate_t *pstate,
@@ -161,17 +168,18 @@ idl_declare(
   idl_scope_t *scope,
   idl_declaration_t **declarationp)
 {
-  idl_declaration_t *entry = NULL;
+  idl_declaration_t *entry = NULL, *last = NULL;
   int (*cmp)(const char *, const char *);
 
   assert(pstate && pstate->scope);
-  cmp = (pstate->flags & IDL_FLAG_CASE_SENSITIVE) ? &strcmp : &idl_strcasecmp;
+  cmp = (pstate->config.flags & IDL_FLAG_CASE_SENSITIVE) ? &strcmp : &idl_strcasecmp;
 
   /* ensure there is no collision with an earlier declaration */
   for (entry = pstate->scope->declarations.first; entry; entry = entry->next) {
     /* identifiers that differ only in case collide, and will yield a
        compilation error under certain circumstances */
     if (cmp(name->identifier, entry->name->identifier) == 0) {
+      last = entry;
       switch (entry->kind) {
         case IDL_SCOPE_DECLARATION:
           /* declaration of the enclosing scope, but if the enclosing scope
@@ -191,13 +199,48 @@ idl_declare(
           if (kind == IDL_MODULE_DECLARATION)
             goto exists;
           goto clash;
+        case IDL_FORWARD_DECLARATION:
+          /* instance declarations cannot occur in the same scope */
+          assert(kind != IDL_INSTANCE_DECLARATION);
+          /* use declarations are solely there to reserve a name */
+          if (kind == IDL_USE_DECLARATION)
+            goto exists;
+          /* multiple forward declarations are legal */
+          if (kind == IDL_FORWARD_DECLARATION && is_same_type(node, entry->node))
+            continue; /* skip forward to search for defintion */
+          if (kind != IDL_SPECIFIER_DECLARATION || !is_same_type(node, entry->node))
+            goto clash;
+          assert(idl_mask(entry->node) & IDL_FORWARD);
+          {
+            idl_forward_t *forward = (idl_forward_t *)entry->node;
+            /* skip ahead to definition for reporting name clash */
+            if (forward->type_spec)
+              continue;
+            /* FIXME: type specifier in forward declarations is not reference
+                      counted. see tree.h for details */
+            forward->type_spec = node;
+          }
+          break;
         case IDL_USE_DECLARATION:
           if (kind == IDL_INSTANCE_DECLARATION)
             goto exists;
           /* fall through */
         case IDL_SPECIFIER_DECLARATION:
+          /* use declarations are solely there to reserve a name */
           if (kind == IDL_USE_DECLARATION)
             goto exists;
+          /* multiple forward declarations are legal. backward declarations,
+             although unusual, are legal too(?) functionally, they can be
+             discarded, but a reference is required for validation */
+          if (kind == IDL_FORWARD_DECLARATION && is_same_type(node, entry->node))
+          {
+            idl_forward_t *forward = node;
+            assert(forward->type_spec == NULL);
+            /* FIXME: type specifier in forward declarations is not reference
+                      counted. see tree.h for details */
+            forward->type_spec = entry->node;
+            break;
+          }
           /* short-circuit on parsing existing annotations */
           if (is_consistent(pstate, node, entry->node))
             goto exists;
@@ -240,7 +283,8 @@ clash:
   switch (kind) {
     case IDL_MODULE_DECLARATION:
     case IDL_ANNOTATION_DECLARATION:
-    case IDL_SPECIFIER_DECLARATION: {
+    case IDL_SPECIFIER_DECLARATION:
+    case IDL_FORWARD_DECLARATION: {
       size_t cnt = 0, len, off = 0;
       const char *sep = "::";
       idl_scoped_name_t *scoped_name = NULL;
@@ -295,9 +339,12 @@ clash:
       break;
   }
 
+  last = entry;
+
 exists:
+  assert(last);
   if (declarationp)
-    *declarationp = entry;
+    *declarationp = last;
   return IDL_RETCODE_OK;
 }
 
@@ -330,7 +377,7 @@ idl_find(
   const idl_name_t *name,
   uint32_t flags)
 {
-  const idl_declaration_t *entry;
+  const idl_declaration_t *entry, *fwd = NULL;
   int (*cmp)(const idl_name_t *, const idl_name_t *);
 
   if (!scope)
@@ -346,7 +393,12 @@ idl_find(
     if (is_annotation(scope, entry) && !(flags & IDL_FIND_ANNOTATION))
       continue;
     if (cmp(name, entry->name) == 0)
-      return entry;
+    {
+      if (entry->kind == IDL_FORWARD_DECLARATION)
+        fwd = entry;
+      else
+        return entry;
+    }
   }
 
   if (!(flags & IDL_FIND_IGNORE_IMPORTS)) {
@@ -357,7 +409,8 @@ idl_find(
     }
   }
 
-  return NULL;
+  assert(!entry);
+  return fwd;
 }
 
 const idl_declaration_t *
@@ -449,7 +502,7 @@ idl_resolve(
 
   if (kind == IDL_ANNOTATION_DECLARATION)
     flags |= IDL_FIND_ANNOTATION;
-  if (pstate->flags & IDL_FLAG_CASE_SENSITIVE)
+  if (pstate->config.flags & IDL_FLAG_CASE_SENSITIVE)
     ignore_case = 0u;
 
   if (scoped_name->absolute)
@@ -542,7 +595,7 @@ idl_scope_t *idl_scope(const void *node)
 idl_declaration_t *idl_declaration(const void *node)
 {
   if (idl_is_declaration(node)) {
-    assert(((const idl_node_t *)node)->declaration);
+    assert(((const idl_node_t *)node)->declaration || idl_mask(node) & IDL_BITMASK);
     return (idl_declaration_t *)((const idl_node_t *)node)->declaration;
   } else {
     assert(!((const idl_node_t *)node)->declaration);
